/*
 * misc.c - miscellaneous routines
 */

#include "config.h"

#include <stdio.h>
#include <string.h>

#if HAS_STDLIB
# include <stdlib.h>
#else
  char *malloc();
#endif /* HAS_STDLIB */

#include "common.h"
#include "yacc.h"

#include "misc.h"



#ifdef MSDOS
#define mem_roundup(x) (x)
#endif

#ifndef MSDOS
#define mem_roundup(x) (((x) + 3) & -4)
#endif



global void error1(fmt, x)
    char *fmt; 
    char *x;
{
    char c;
    
    printf("%s %d: ", filename, lineno);
    while ((c = *fmt++) != '\0') {
	if (c == '%' && *fmt != '\0') {
	    switch (c = *fmt++) {
	    case 's':
		fputs(x, stdout);
		break;
	    default:
		putchar(c);
		break;
	    }
	} else {
	    putchar(c);
	}
    }
    putchar('\n');
}


global void error(fmt)
    char *fmt;
{
    error1(fmt, NULL);
}


/* Report error and exit */
global void die(msg)
    char *msg;
{
    error(msg);
    exit(1);
}


/* Report error and exit */
global void die1(msg, str)
    char *msg;
    char *str;
{
    error1(msg, str);
    exit(1);
}

/* Memory allocation */



#if LARGEMEM

#define ALLOCTHRESH 128
#define CHUNKSIZE 4096

static char *alloc_cp;
static int alloc_left;
static long size_allocated;

#endif /* LARGEMEM */


#if !LARGEMEM

# define MEMSIZE (40 * 1024)

static char bigmem_top[MEMSIZE];
#define bigmem_bottom (bigmem_top + MEMSIZE)
static char *alloc_cp = bigmem_top;
static char *talloc_cp = bigmem_bottom;

#endif /* !LARGEMEM */


/* Allocate persistent object */
global void *alloc(s)
    unsigned s;
{
#if LARGEMEM
    char *ap;

    s = mem_roundup(s);
    size_allocated += s;
    if (s >= ALLOCTHRESH) {
	if ((ap = malloc(s)) == NULL)
	    die("out of memory");
    } else {
	if (s > alloc_left) {
	    if ((alloc_cp = malloc(CHUNKSIZE)) == NULL)
		die("out of memory");
	    alloc_left = CHUNKSIZE;
	}
	ap = alloc_cp;
	alloc_cp += s;
	alloc_left -= s;
    }
    memset(ap, 0, s);
    return ap;
#endif /* LARGEMEM */

#if !LARGEMEM
    char *ap;

    s = mem_roundup(s);
    ap = alloc_cp;
    if (s > talloc_cp - ap)
	die("out of memory");
    alloc_cp += s;
    memset(ap, 0, s);
    return ap;
#endif /* !LARGEMEM */
}


/* Allocate temporary object (freeable by release()) */
global void *talloc(s)
    unsigned s;
{
#if LARGEMEM
    return alloc(s);
#endif /* LARGEMEM */

#if !LARGEMEM
    s = mem_roundup(s);
    if (s > talloc_cp - alloc_cp)
	die("out of memory");
    talloc_cp -= s;
    memset(talloc_cp, 0, s);
    return talloc_cp;
#endif /* !LARGEMEM */
}


/* Release temporary object allocated by talloc() */
global void release()
{
#if !LARGEMEM
    talloc_cp = bigmem_bottom;
#endif /* !LARGEMEM */
}


global long bytes_allocated()
{
#if LARGEMEM
    return size_allocated;
#endif /* LARGEMEM */

#if !LARGEMEM
    return (bigmem_bottom - talloc_cp) + (alloc_cp - bigmem_top);
#endif /* !LARGEMEM */
}


global void show_mem_usage()
{
#if LARGEMEM
    fprintf(stderr, "(%ld)\n", size_allocated);
#endif /* LARGEMEM */

#if !LARGEMEM
    fprintf(stderr, "(%u/%u/%u)\n", alloc_cp - bigmem_top,
	    talloc_cp - alloc_cp, bigmem_bottom - talloc_cp);
#endif /* !LARGEMEM */
}
