/*
 * token.c - lexical analyzer for yacc
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#if HAS_STDLIB
# include <stdlib.h>
#endif /* HAS_STDLIB */

#include "common.h"
#include "yacc.h"

#include "token.h"
#include "misc.h"


#ifndef global
/* return values of function gettoken */
#define	NAME		'a'
#define	NUMBER		'0'
#define	MARK		0x100
#define	BEGININC	0x101
#define	ENDINC		0x102
#define	TOKEN		0x103
#define	LEFT		0x104
#define	RIGHT		0x105
#define	NONASSOC	0x106
#define	PRECTOK		0x107
#define	TYPE		0x108
#define	UNION		0x109
#define	START		0x10a

#define EXPECT		0x10c

#endif /* global */



#define	MAXBACK	100
static char back_buff[MAXBACK];
static char *backp = back_buff;


global int get()
{
    int c;
    
    if (backp > back_buff)
	return (*--backp);
    if ((c = getc(ifp)) == '\n')
	lineno++;
    return (c);
}


global void unget(c)
    int c;
{
    if (c == EOF)
	return;
    if (backp >= back_buff + MAXBACK)
	error("too many unget");
    *backp++ = c;
}



global void ungettok(s)
    char *s;
{
    char *p;
    
    for (p = s + strlen(s); --p >= s;)
	unget(*p);
}




void putifopen(c, fp, fp2)
    int c;
    FILE *fp;
    FILE *fp2;
{
    if (fp != NULL)
	putc(c, fp);
    if (fp2 != NULL)
	putc(c, fp2);
}


int skipwhite(fp, fp2)
    FILE *fp;
    FILE *fp2;
{
    int c;
    
    for (;;) {
	while (c = get(), isspace(c))
	    putifopen(c, fp, fp2);
	if (c != '/')
	    return (c);
	if ((c = get()) != '*' && c != '/') {
	    unget(c);
	    return ('/');
	}
	/* skip comments */
	if (c == '*') {
	    putifopen('/', fp, fp2);
	    putifopen('*', fp, fp2);
	    for (;;) {
		if ((c = get()) == '*') {
		    if ((c = get()) == '/')
			break;
		    unget(c);
		}
		if (c == EOF) {
		    error("missing */");
		    exit(1);
		}
		putifopen(c, fp, fp2);
	    }
	    putifopen('*', fp, fp2);
	    putifopen('/', fp, fp2);
	} else {
	    /* do not copy // comment to output */
	    while ((c = get()) != '\n' && c != EOF)
		;
	    putifopen('\n', fp, fp2);
	}
    }
}



#define	issymch(c) (isdigit(c) || isalpha(c) || c == '_' || c == '.')


global int copying_gettoken(buff, fp, fp2)
    char *buff;
    FILE *fp;
    FILE *fp2;
{
    int c, tag;
    char *p;
    
    c = skipwhite(fp, fp2);
    buff[0] = '\0';
    if (c == EOF)
	return (EOF);
    p = buff;
    tag = c;
    if (c == '%') {
	*p++ = c;
	c = get();
    }
    if (issymch(c)) {
	while (issymch(c)) {
	    *p++ = c;
	    c = get();
	}
	unget(c);
	tag = isdigit(tag) ? NUMBER : NAME;
    } else if (c == '\'' || c == '"') {
	*p++ = c;
	while ((*p++ = c = get()) != tag) {
	    if (c == EOF || c == '\n') {
		error("missing '");
		break;
	    }
	    if (c == '\\')
		*p++ = get();
#if MSKANJI
	    if (iskanji(c)) {
		c = get();
		if (!iskanji2(c)) {
		    error("bad kanji code");
		}
		*p++ = c;
	    }
#endif
	}
    } else
	*p++ = c;
    *p++ = '\0';
    if (buff[0] == '%') {
	/* check keyword */
	if (strcmp(buff, "%%") == 0)
	    tag = MARK;
	else if (strcmp(buff, "%{") == 0)
	    tag = BEGININC;
	else if (strcmp(buff, "%}") == 0)
	    tag = ENDINC;
	else if (strcmp(buff, "%token") == 0)
	    tag = TOKEN;
	else if (strcmp(buff, "%term") == 0)	/* old feature */
	    tag = TOKEN;
	else if (strcmp(buff, "%left") == 0)
	    tag = LEFT;
	else if (strcmp(buff, "%right") == 0)
	    tag = RIGHT;
	else if (strcmp(buff, "%nonassoc") == 0)
	    tag = NONASSOC;
	else if (strcmp(buff, "%prec") == 0)
	    tag = PRECTOK;
	else if (strcmp(buff, "%type") == 0)
	    tag = TYPE;
	else if (strcmp(buff, "%union") == 0)
	    tag = UNION;
	else if (strcmp(buff, "%start") == 0)
	    tag = START;
	else if (strcmp(buff, "%expect") == 0)
	    tag = EXPECT;
	else {
	    if (!issymch(buff[1])) {
		ungettok(buff + 1);
		buff[1] = '\0';
	    }
	    tag = '%';
	}
    }
    return (tag);
}


global int gettoken(buff)
    char *buff;
{
    return (copying_gettoken(buff, NULL, NULL));
}


global int peekch()
{
    int c;
    
    c = skipwhite(NULL, NULL);
    unget(c);
    return (c);
}


global int getnum()
{
    int c, v, sign;
    
    v = 0;
    if ((sign = get()) == '-')
	c = get();
    else
	c = sign;
    while (isdigit(c)) {
	v = v * 10 + (c - '0');
	c = get();
    }
    unget(c);
    return (sign == '-' ? -v : v);
}
