/*
 * yacc.c - LALR parser generator
 *
 * Copyright (C) 1987,1989,1992,1993 by MORI Koichiro.
 * All Rights Reserved.
 */

static char sccsid[] = "@(#)yacc.c 1993/11/12 20:54 3.08; by MORI Koichiro";

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#if HAS_STDLIB
# include <stdlib.h>
#else
 char *getenv();
#endif /* HAS_STDLIB */

#include "common.h"
#include "misc.h"
#include "token.h"

#include "grammar.h"
#include "lalr.h"
#include "table.h"



global bool lflag;
global bool debug;

global FILE *ifp, *ofp, *vfp, *hfp, *afp;
global int lineno
	= 1;
global char *filename;

global int expected_srconf;

global char *pspref;


static char *progname = "kmyacc";

#ifdef MSDOS
# define MAXPATHLEN 90
#else
# define MAXPATHLEN 200
#endif /* MSDOS */


/* Print usage of this program */
void usage()
{
    fprintf(stderr, "KMyacc - parser generator ver 3.08\n");
    fprintf(stderr, "Copyright (C) 1987-1989,1992,1993 by MORI Koichiro\n\n");
    fprintf(stderr, "Usage: %s [-dvlta] [-b y] [-p yy] [-m model] grammar.y\n", progname);
}


/* Open file fn with mode; exit if fail */
FILE *efopen(fn, mode)
    char *fn;
    char *mode;
{
    FILE *fp;

    if (strcmp(fn, "-") == 0) {
	switch (*mode) {
	case 'r': return stdin;
	case 'w': return stdout;
	}
    }
    if ((fp = fopen(fn, mode)) == NULL) {
	fprintf(stderr, "%s: ", progname);
	perror(fn);
	exit(1);
    }
    return (fp);
}


/* Close file fp and exit if error */
void efclose(fp)
    FILE *fp;
{
    if (ferror(fp) || fclose(fp)) {
	fprintf(stderr, "%s: can't close\n", progname);
	exit(1);
    }
}


#ifdef MSDOS
# define FNPREF "ytab"
# define C_SUFFIX ".c"
# define H_SUFFIX ".h"
# define OUTPREF "y"
# define OUT_SUFFIX ".out"
# define ACTTEMP "kmXXXXXX"
#else
# define FNPREF "y"
# define C_SUFFIX ".tab.c"
# define H_SUFFIX ".tab.h"
# define OUTPREF "y"
# define OUT_SUFFIX ".output"
# define ACTTEMP "/tmp/kmXXXXXX"
# define remove unlink
#endif /* MSDOS */


/* Entry point */
void main(argc, argv)
    int argc;
    char *argv[];
{
    static char acttemp[] = ACTTEMP;
    char fn[MAXPATHLEN];
    char buf[80];
    char *parserfn;
    int c;
    char *p;
    FILE *pfp;
    int plineno;

    bool vflag = NO;
    bool dflag = NO;
    bool tflag = NO;
    bool aflag = NO;
    char *fnpref = FNPREF;
    char *outpref = OUTPREF;

    if ((parserfn = getenv("KMYACCPAR")) == NULL)
	parserfn = PARSERFN;

#ifndef MSDOS
    progname = *argv;
#endif /* !MSDOS */

    while (++argv, --argc != 0 && argv[0][0] == '-') {
	for (p = argv[0] + 1; *p; ) {
	    switch (*p++) {
	    case 'd':
		dflag = YES;
		break;
	    case 'x':
		debug = YES;
	    /* fall thru */
	    case 'v':
		vflag = YES;
		break;
	    case 'l':
		lflag = YES;
		break;
	    case 't':
		tflag = YES;
		break;
	    case 'a':
		aflag = YES;
		break;
	    case 'b':
		if (*p == '\0') {
		    if (--argc <= 0)
			goto boo;
		    fnpref = *++argv;
		} else {
		    fnpref = p;
		    p = "";
		}
		outpref = fnpref;
		break;
	    case 'p':
		if (*p == '\0') {
		    if (--argc <= 0)
			goto boo;
		    pspref = *++argv;
		} else {
		    pspref = p;
		    p = "";
		}
		break;
	    case 'm':
		if (*p == '\0') {
		    if (--argc <= 0)
			goto boo;
		    parserfn = *++argv;
		} else {
		    parserfn = p;
		    p = "";
		}
		break;
	    default:
	    boo:
		usage();
		exit(1);
	    }
	}
    }
    if (argc != 1) {
	usage();
	exit(1);
    }
    filename = argv[0];
    ifp = efopen(argv[0], "r");
    ofp = efopen(strcat(strcpy(fn, fnpref), C_SUFFIX), "w");
    if (dflag)
	hfp = efopen(strcat(strcpy(fn, fnpref), H_SUFFIX), "w");

    /* Copy parser - 1/3 */
    pfp = efopen(parserfn, "r");
    plineno = 1;
    if (!lflag)
	fprintf(ofp, "#line 1 \"%s\"\n", parserfn);
    while ((c = getc(pfp)) != EOF) {
	if (c == '$') {
	    c = getc(pfp);
	    if (c != 'y')
		break;
	    c = getc(pfp);
	    if (c != 'y') {
		fprintf(stderr, "%s: illegal $yy\n", parserfn);
		exit(1);
	    }
	    p = buf;
	    while ((c = getc(pfp)) != EOF && c != '\n')
		*p++ = c;
	    *p = '\0';
	    if (pspref)
		fprintf(ofp, "#define yy%s %s%s\n", buf, pspref, buf);
	    plineno++;
	} else {
	    putc(c, ofp);
	    if (c == '\n')
		plineno++;
	}
    }
    if (c != 'D') {
	fprintf(stderr, "%s: No $D sign\n", parserfn);
	exit(1);
    }
    
    /* Read declaration section */
    do_declaration();

    if (hfp) efclose(hfp);

    if (tflag) {
	fprintf(ofp, "\n#ifndef YYDEBUG\n");
	fprintf(ofp, "# define YYDEBUG 1\n");
	fprintf(ofp, "#endif /* YYDEBUG */\n");
    }

    if (aflag)
	fprintf(ofp, "\n#define YYACTFUN 1\n\n");

    /* Copy parser - 2/3 */
    if (!lflag)
	fprintf(ofp, "#line %d \"%s\"\n", plineno, parserfn);
    while ((c = getc(pfp)) != EOF && c != '$') {
	putc(c, ofp);
	if (c == '\n')
	    plineno++;
    }
    if (c != '$' || (c = getc(pfp)) != 'T' && c != 'A') {
	fprintf(stderr, "%s: No $T sign", parserfn);
	exit(1);
    }
    if (c == 'A')
	aflag = YES;
    if (!aflag) {
	if (mktemp(acttemp) == NULL) {
	    perror("can't create temporary file");
	    exit(1);
	}
	afp = efopen(acttemp, "w");
    } else {
	afp = NULL;
    }
    /* Second $ is replaced by parser tables (and action routines) */

    /* read grammar */
    do_grammar();
#if 0
 fprintf(stderr, "grammar read. ");
 show_mem_usage();
#endif

    if (afp) efclose(afp);

    if (worst_error)
	goto final;

    insertline(ofp);
    while ((c = get()) != EOF)
	putc(c, ofp);
    fclose(ifp);

    if (vflag)
	vfp = efopen(strcat(strcpy(fn, outpref), OUT_SUFFIX), "w");

    /* compute LALR(1) states & actions */
    comp_lalr();

    /* generate parsing table */
    gentable();
#if 0
 fprintf(stderr, "gentable done. ");
 show_mem_usage();
#endif

    if (vfp) efclose(vfp);

    /* Copy parser - 3/3 */
    if (!lflag)
	fprintf(ofp, "#line %d \"%s\"\n", plineno, parserfn);
    while ((c = getc(pfp)) != EOF) {
	if (c == '$' && getc(pfp) == 'A') {
	    if (!aflag) {
		/* Copy actions as switch-body */
		afp = efopen(acttemp, "r");
		while ((c = getc(afp)) != EOF)
		    putc(c, ofp);
		fclose(afp);
		if (!lflag)
		    fprintf(ofp, "#line %d \"%s\"\n", plineno, parserfn);
	    }
	} else {
	    putc(c, ofp);
	    if (c == '\n')
		plineno++;
	}
    }

 final:
    fclose(pfp);
    efclose(ofp);
    if (!aflag)
	remove(acttemp);
	
    exit(worst_error);
}
